/*****************************************************************************
 * Program Name:  Storage Management Services (Example Engine)
 *
 * Filename:	  de1.c
 *
 * Date Created:  March 14, 1991
 *
 * Version:		  1.0
 *
 * Files used:	  de.h
 *
 * Functions Defined In This File:   ListTSAs
 *				    ListTSs
 *				    PrintError
 *				    ListSelectionStrings
 *				    ListScanTypeStrings
 *				    EstablishConnection
 *				    BuildSelectionList
 *				    DisplaySelectionList
 *				    FreeSelectionList
 *				    Backup
 *				    Restore
 *				    EraseFile
 *				    RenameFile
 *				    main
 *
 *
 * Date Modified: 
 *
 * Modifications:
 *
 *
 *
 *
 * Comments:	    DE is a sample SME
 *
 * (C) Unpublished Copyright of Novell, Inc.  All Rights Reserved.
 *
 * No part of this file may be duplicated, revised, translated, localized or
 * modified in any manner or compiled, linked or uploaded or downloaded to or
 * from any computer system without the prior written consent of Novell, Inc.
 ****************************************************************************/

 /***For NLM Conversion***/
 /****************************************************************************
 *	NOTE: Comments are included with this code that describe how to
 *		convert this demonstration engine to an NLM. To make an NLM 
 *		well behaved, control should be passed to other processes on
 *		the server at least every 150 milliseconds (386 @ 25Mhz).
 *		To accomodate this, calls to the ThreadSwitch function (CLIB) 
 *		should be appropriately scattered through the code. 
 *		ThreadSwitches should be considered for any sections of
 *		code, such as loops, which could execute for more 
 *		than 150 milliseconds. CLIB routines that block should already
 *		pass control to other threads so you won't have to worry about
 *		these.
 *
 *		To find the changes need for converting this demonstration 
 *		to an NLM, search for the string "NLM" in this code.
 ****************************************************************************/

/*** FOR NLM CONVERSION ***
* remove #include <nwcalls.h>
****************************/
#include <dos.h>
#include <sms.h>
#include <nwcalls.h>
#include "de.h"

#define bdt "BuIlDdAtEaNdTiMe="

extern unsigned _stklen = 4*1024;

char *_cdt_ = bdt __DATE__ " @ "  __TIME__;
int _cdt_o_ = sizeof(bdt) - 1;

UINT32 connection = 0;
NWSM_SELECTION_LIST *selectionList = NULL;

/****************************************************************************/
/*********************************************************************
* NAME: ListTSAs
*
* WHAT: Display a list of the TSAs visible under the target SMDR.
*
* SMS-SPECIFIC ROUTINES CALLED: NWSMListTSAs (TSAPI)
*				NWSMFreeNameList (TSAPI)
*
**********************************************************************/
void ListTSAs(char *target)
{
	NWSM_NAME_LIST *list=NULL;

	clrscr();
	strcat(target,".*");
	puts("Visible Target Service Agents:");
	NWSMListTSAs(target,&list);
	DisplayList(list);
	NWSMFreeNameList(&list);
}
/****************************************************************************/

/*********************************************************************
* NAME: ListSMDRs
*
* WHAT: Display a list of the SMDRs visible on the network.
*
* SMS-SPECIFIC ROUTINES CALLED: NWSMListSMDRs (TSAPI)
*				NWSMFreeNameList (TSAPI)
*
**********************************************************************/
void ListSMDRs(void)
{
	NWSM_NAME_LIST *list=NULL;

	clrscr();
	puts("Visible Storage Management Data Requestors:");
	NWSMListSMDRs("*",&list);
	DisplayList(list);
	NWSMFreeNameList(&list);
}
/****************************************************************************/

/*********************************************************************
* NAME: ListTSs
*
* WHAT: Display a list of the Target Services on the network.
*
* SMS-SPECIFIC ROUTINES CALLED: NWSMTSListTargetServices (TSAPI)
*				NWSMFreeNameList (TSAPI)
*
**********************************************************************/
void ListTSs(void)
{
	NWSM_NAME_LIST *list=NULL;

	clrscr();
	puts("Target Services:");
	NWSMTSListTargetServices(connection, "*", &list);
	DisplayList(list);
	NWSMFreeNameList(&list);
	pause();
}
/****************************************************************************/

/*********************************************************************
* NAME: PrintError
*
* WHAT: Convert Error Return (CCODE) to a string and display it.
*
* SMS-SPECIFIC ROUTINES CALLED: NWSMConvertError (SMS Utility API)
*
**********************************************************************/
void PrintError(CCODE ccode, char *routine, UINT16 line)
{
	BUFFER buffer[300];

	NWSMConvertError(connection, ccode, buffer);
	if (line)
		printf("\n\n%s:  (line = %u)\n   \"%s\"\n\n", routine, line, buffer);

	else 
		printf("\n\n%s:\n   \"%s\"\n\n", routine, buffer);
}
/****************************************************************************/

/*********************************************************************
* NAME: ListSelectionStrings
*
* WHAT: Display a list of selection strings used to include/exclude
*	data sets to be scanned on the Target Service. Selection types
*	are used conjunction with scan types to control data set 
*	scanning.
*
* SMS-SPECIFIC ROUTINES CALLED: NWSMTSGetTargetSelectionTypeStr (SIDR)
*
**********************************************************************/
void ListSelectionStrings(void)
{
	CCODE ccode = 0;
	char string[256];
	UINT8 typeNumber;

	puts("Selection Strings: ");
	for (typeNumber = 1; !ccode; typeNumber++)
	{
		ccode = NWSMTSGetTargetSelectionTypeStr(connection, typeNumber, string,"++");
		printf("  %s\n", string);
	}

	pause();
}
/****************************************************************************/

/*********************************************************************
* NAME: ListScanTypeStrings
*
* WHAT: Display a list of scan type strings used to include/exclude
*	data sets to be scanned on the Target Service. Scan types are
*	used in conjunction with selection types to control data set
*	scanning.
*
* SMS-SPECIFIC ROUTINES CALLED: NWSMTSGetTargetScanTypeString (SIDR)
*
**********************************************************************/
void ListScanTypeStrings(void)
{
	CCODE ccode = 0;
	char string[256];
	UINT8 typeNumber;
	UINT32 req, dis;

	puts("ScanType Strings: ");
	for (typeNumber = 0; !ccode; typeNumber++)
	{
		ccode = NWSMTSGetTargetScanTypeString(connection, typeNumber, string, &req, &dis);
		printf("  %s\n", string);
	}

	pause();
}
/****************************************************************************/

/*********************************************************************
* NAME: EstablishConnection
*
* WHAT: Establish a connection between the SME and the Target Service
*	Agent on the target to be scanned,backed up or restored.
*
* SMS-SPECIFIC ROUTINES CALLED: NWSMConnectToTSA (TSAPI)
*				NWSMTSConnectToTargetService (TSAPI)
*
**********************************************************************/
CCODE EstablishConnection(void)
{
	CCODE ccode;
	char target[100], user[100], password[40];
	NWSM_NAME_LIST *services = NULL;
	NWSM_NAME_LIST *p;

	ListSMDRs();
	printf("SMDR:       "); gets(target);
	printf("\n\n\n           Please wait...\n\n");

	ListTSAs(target);
	printf("TSA:       "); gets(target);
	printf("\n\n\n           Please wait...\n\n");

	if ((ccode = NWSMConnectToTSA(target, &connection)) isnt 0)
	{
		puts("Failure to connect to Target Service Agent.");
		connection = 0;
		goto Return;
	}

	ccode = NWSMTSListTargetServices(connection, "*", &services);
	if (ccode) {
		goto Return;
	}

	printf("\n\tTarget Service(s): ");
	for (p = services; p!=NULL; p = p->next) {
		if(p->name) puts((char *)p->name);
	}

	printf("\n\nWhich target service do you want to connect to?  "); gets(target);
	printf("User name: "); getun(user);
	printf("Password:  "); getpw(password);
	ccode = NWSMTSConnectToTargetService(&connection, target, user, password);
	if(ccode){
	    if (connection)
	    {
		    NWSMReleaseTSA(&connection);
	    }
	}


Return:
	if(services) NWSMFreeNameList(&services);
	return (ccode);
}
/****************************************************************************/


/*********************************************************************
* NAME: BuildSelectionList
*
* WHAT: Build a list for including/excluding data sets in/from a scan
*	of a Target Service.
*	The user is prompted for name space type, selection type, and
*	selection item for each element in the list.
*
* SMS-SPECIFIC ROUTINES CALLED: GetNameSpaceType (DE2.C)
*				GetSelectionType (DE2.C)
*				NWSMPutFirstName (SMS UTILITY API)
*				NWSMPutNextName (SMS UTILITY API)
*				NWSMCloseName (SMS UTILITY API)
*
**********************************************************************/
void BuildSelectionList(void)
{
	BUFFER buffer[2000];
	CCODE ccode;
	STRING_BUFFER *firstSep=NULL, *nextSep=NULL;
	NWBOOLEAN firstTime = TRUE;
	NWBOOLEAN reverseOrder = FALSE;
	UINT32 nameHandle = 0, nameSpaceType, selectionType;

	if (selectionList)
	{
		free(selectionList);
		selectionList = NULL;
	}

	clrscr();

	puts("Enter selection element:");
	loop
	{
		nameSpaceType = GetNameSpaceType();
		if (nameSpaceType == 100 )
			goto Return;

		ccode = NWSMTSGetNameSpaceTypeInfo(connection, nameSpaceType, &reverseOrder, &firstSep, &nextSep);
		if (ccode != 0)
			{
			printf("\n\a\aSomething's wrong here\n");
			nameSpaceType = 100;
			goto Return;
			}

		if ((selectionType = GetSelectionType()) is 100)
			goto Return;

		printf("\n\nEnter selection item: "), gets(buffer);
		if (firstTime)
		{
			ccode = NWSMPutFirstName((void *)&selectionList, nameSpaceType,
					selectionType, reverseOrder, firstSep->string, nextSep->string, buffer, &nameHandle);
			firstTime = FALSE;
		}
		else
			ccode = NWSMPutNextName((void *)&selectionList, &nameHandle, nameSpaceType,
					selectionType, reverseOrder, firstSep->string, nextSep->string, buffer);

		if (ccode)
		{
			printf("\nNWSMPut%sName = 0x%X\n", firstTime ? "First":"Next", ccode);
			goto Return;
		}

		clrscr();
		NWSMFreeString(&firstSep);
		NWSMFreeString(&nextSep);
	
	}

Return:
	if (!firstTime)
		NWSMCloseName(&nameHandle);
}
/****************************************************************************/

/*********************************************************************
* NAME: DisplaySelectionList
*
* WHAT: Display the list used for selecting data sets from the target
*	service. This list is created by the BuildSelectionList function.
*
* SMS-SPECIFIC ROUTINES CALLED: NWSMGetFirstName (SMS UTILITY API)
*				SetNameSpaceType (DE2.C)
*				SetSelectionType (DE2.C)
*				NWSMGetNextName (SMS UTILITY API)
*				NWSMCloseName (SMS UTILITY API)
*
**********************************************************************/
void DisplaySelectionList(void)
{
	CCODE ccode;
	NWSM_DATA_SET_NAME name;
	UINT32 handle = 0;
	char string1[50];
	char string2[128];

	if ((ccode = NWSMGetFirstName((void *)selectionList, &name, &handle)) isnt 0)
	{
		printf("\nNWSMGetFirstName = 0x%X\n", ccode);
		goto Return;
	}

	clrscr();
	puts("SelectionList:\n");
	do
	{
		SetNameSpaceType(name.nameSpaceType, string1);
		SetSelectionType(name.selectionType, string2);
		printf("%s, %s:\n  \"%s\"\n\n", string1, string2, name.name);

		if ((ccode = NWSMGetNextName(&handle, &name)) isnt 0 and ccode isnt NWSMUT_NO_MORE_NAMES)
			printf("\nNWSMGetNextName = 0x%X\n", ccode);
	} while (!ccode);

	puts("");
	pause();

Return:
	if (handle and (ccode = NWSMCloseName(&handle)) isnt 0)
		printf("\nNWSMCloseName = 0x%X\n", ccode);
}
/****************************************************************************/

/*********************************************************************
* NAME: FreeSelectionList
*
* WHAT: Free the memory allocated for the data set selection list.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
void FreeSelectionList(void)
{
	if (selectionList)
	{
		free(selectionList);
		selectionList = NULL;
	}
}
/****************************************************************************/

NWSM_SCAN_CONTROL scanControl =
{
	/* bufferSize */                         sizeof(NWSM_SCAN_CONTROL),
	/* scanControlSize */                    sizeof(NWSM_SCAN_CONTROL) - sizeof(UINT16) - sizeof(UINT8),
	/* scanType */                           0,
	/* firstAccessDateAndTime */             0,
	/* lastAccessDateAndTime */              0,
	/* firstCreateDateAndTime */             0,
	/* lastCreateDateAndTime */              0,
	/* firstModifiedDateAndTime */           0,
	/* lastModifiedDateAndTime */            0,
	/* firstArchivedDateAndTime */           0,
	/* lastArchivedDateAndTime */            0,
	/* returnChildTerminalNodeNameOnly */    TRUE,
	/* parentsOnly */                        FALSE,
	/* childrenOnly */                       FALSE,
	/* createSkippedDataSetsFile */          TRUE,
	/* generateCRC */                        FALSE,
	/* returnHardLinksInDataSet */			  FALSE,
	/* reserved */                           { 0 },
	/* scanChildNameSpaceType */             NWSM_ALL_NAME_SPACES,
	/* returnNameSpaceType */                NWSM_ALL_NAME_SPACES,
	/* callScanFilter */                     FALSE,
	/* otherInformationSize */               0
};
/****************************************************************************/

NWSM_SCAN_INFORMATION *scanInfo;
BUFFER dataBuffer[BUFFER_SIZE];

/*********************************************************************
* NAME: Backup
*
* WHAT: This function controls the overall process of backing up the
*	Target Service. A complimentary function (Restore) exists for
*	restoring data to a Target Service.
*
* SMS-SPECIFIC ROUTINES CALLED: NWSMPutOneName (SMS UTILITY API)
*				NWSMTSScanDataSetBegin (TSAPI)
*				NWSMGetDataSetName (SMS UTILITY API)
*				NWSMTSScanDataSetEnd (TSAPI)
*				NWSMTSOpenDataSetForBackup (TSAPI)
*				NWSMTSCloseDataSet (TSAPI)
*				NWSMTSReadDataSet (TSAPI)
*				NWSMTSScanNextDataSet (TSAPI)
*
**********************************************************************/
void Backup(void)
{
	CCODE ccode;
	char myName[48];
	char *prompt;
	char option;
	int backup = TRUE, command, resource, restart = TRUE;
	int resourceListCount=1;
	int x, y;
	int index;
	UINT32 total, utotal;

	NWSM_DATA_SET_NAME dataSetName;
	NWSM_DATA_SET_NAME_LIST *resourceName, *dataSetNameList;
	NWSM_NAME_LIST *serviceResourceList;
	NWSM_NAME_LIST *saveFirstResourceList;
	UINT16 resourceCount, readSize;
	UINT32 bytesRead, dataHandle, sequence;

	serviceResourceList = saveFirstResourceList = NULL;


Loop:
	utotal = 0;
	resourceName = dataSetNameList = NULL;
	scanInfo = NULL;
	resourceCount = 0;
	readSize = BUFFER_SIZE;
	sequence = 0L;

/*********************************************************************
*	Prompt the user for a backup or scan command: Backup Resource,
*		Backup Subdirectories, Scan Resource, Scan Subdirectories,
*		Save Backup Data (toggle on/off).
**********************************************************************/
	switch (command = GetBCommand(backup, &prompt, &resource))
	{
	case 'T':
		backup = !backup;
		goto Loop;

/*********************************************************************
*	Prompt the user for the name of the resource or subdirectory to
*	scan. This resource/subdirectory is added to a list for backup
*	or scan.
**********************************************************************/
	case 'B':
	case 'S':
		printf("\nEnter %s to %s: ", prompt, command is 'B' ? "backup" : "scan");
		if (resource)
			{
			resourceListCount = 1;
			NWSMTSListTSResources(connection, &serviceResourceList);
			saveFirstResourceList = serviceResourceList;
			while (serviceResourceList != NULL)
				{
				printf("\n\t%d: %s\n", resourceListCount, serviceResourceList->name);
				serviceResourceList = serviceResourceList->next;
				resourceListCount++;
				}
			printf("\nPlease enter selection: ");
			x=wherex();
			y=wherey();
			option=getch();
			index=option-'0';
			while ((index < 1) || (index >= resourceListCount))
				{
				printf("\a\a");
				gotoxy(x,y);
				option = getch();
				index = option - '0';
				}

			serviceResourceList = saveFirstResourceList;
			for (resourceListCount=1; resourceListCount < index; resourceListCount++)
				{
				serviceResourceList = serviceResourceList->next;
				}
			strcpy(myName, (char *)serviceResourceList->name);


			ccode = NWSMPutOneName((void HUGE **)&resourceName, NWSM_TSA_DEFINED_RESOURCE_TYPE, 0L, 0,
				NULL, NULL, myName);
			CloseDEFiles();  /**  This needs to be done so that you can still do a backup
										after scanning. ***/

			}

		else
			{
			gets(myName);
			ccode = NWSMPutOneName((void HUGE **)&resourceName, DOSNameSpace, 0L, 0,":", "/", myName);
			}
		break;

	default:
		restart = FALSE;
		goto Terminate;
	}

	if (ccode)
	{
		printf("\nNWSMPutOneName:   ccode = %lu\n", ccode);
		restart = FALSE;
		goto Terminate;
	}

/*********************************************************************
*	Begin scanning for the selected data sets on the Target Service.
**********************************************************************/
	if ((ccode = NWSMTSScanDataSetBegin(connection, resourceName, &scanControl, selectionList, &sequence, &scanInfo, &dataSetNameList)) isnt 0)
	{
		PrintError(ccode, "NWSMTSScanDataSetBegin", 0);
		goto Terminate;
	}

	clrscr();

/*********************************************************************
*	For this demonstration engine, the backup data is placed in
*	DOS files instead of tape(or other backup media). Here we open
*	the files.
**********************************************************************/
	if (backup)
		OpenDEFiles('w');

	StartTimer();

/*********************************************************************
*	This loop completes the scan of the selected data sets on the
*	Target Service. The data set name is displayed to the user
*	as it is scanned.
**********************************************************************/

	while (!ccode)
	{
		CCODE error;
		int x, y;

		resourceCount++;
		NWSMGetDataSetName(dataSetNameList, 0L, &dataSetName);
		status(scanInfo->parentFlag ? "%-60s":"  %-58s", dataSetName.name);

/*********************************************************************
*	If the scan option was selected and the user presses any key,
*	terminate the scan.
*
*	NOTE: To clean up properly, NWSMTSScanDataSetEnd must be called
*	if ending a scan prematurely. See NWSMTSScanDataSetEnd in the
*	DR API document for more details.
**********************************************************************/
		if (command is 'S')
		{
			puts("");
			if (kbhit())
			{
				NWSMTSScanDataSetEnd(connection, &sequence, &scanInfo, &dataSetNameList);
				goto Terminate;
			}
		}

/*********************************************************************
*	For the Backup option, each data set returned by the scan
*	is opened for backup. If the open fails, this data set is skipped
*	and the next one is scanned.
*	If the open succeeds a check is made for a non-zero data handle
*	returned by the open. If the data handle is non-zero, read the 
*	data set until complete or until the occurrence of an error.
*	If the option to save the backup data was selected, write the
* 	data to a file on disk.
**********************************************************************/
		else
		{
			UINT32 mode;

			total = 0;
			mode = NWSM_USE_LOCK_MODE_IF_DW_FAILS;
			if ((error = NWSMTSOpenDataSetForBackup(connection, sequence, mode, &dataHandle)) isnt 0)
			{
				PrintError(error, "NWSMTSOpenDataSetForBackup", 0);
				goto NEXT;
			}

			else
				status("Open ");

			if (!dataHandle)
				goto NEXT;

			status("Read");
			y = wherey(), x = wherex();


			while (!error)
			{
				gotoxy(x, y);

/*********************************************************************
*	If the user presses any key - terminate the scan/backup.
**********************************************************************/
				if (kbhit())
				{
				    puts("");
				    if(ccode==0)
					NWSMTSScanDataSetEnd(connection,
					&sequence, &scanInfo, &dataSetNameList);
				    NWSMTSCloseDataSet(connection, &dataHandle);
				    utotal += total;
				    goto Terminate;
				}

/**************************************************************************
*			Read the Data Set 
**************************************************************************/
				if ((error = NWSMTSReadDataSet(connection, dataHandle, readSize, &bytesRead, dataBuffer)) is 0)
				{

/**************************************************************************
* 		Here's where we write the backup data to files 
**************************************************************************/
					if (backup and WriteDEFiles((BUFFER *)dataBuffer, (UINT32)bytesRead,
							scanInfo->parentFlag, scanControl.returnChildTerminalNodeNameOnly))
						break;
				}

				else if (error isnt NWSMTS_NO_MORE_DATA)
					PrintError(error, "NWSMTSReadDataSet", 0);

				total += bytesRead;
				status(" %-8lu ", total);
			/*	save bytesRead bytes from dataBuffer */

				if (bytesRead < readSize)
					break;
			}

			if ((error = NWSMTSCloseDataSet(connection, &dataHandle)) isnt 0)
				PrintError(error, "NWSMTSCloseDataSet", 0);

			puts("");
		}

/*********************************************************************
*	Increment the count of the total number of bytes saved in
*	this backup session.
**********************************************************************/
NEXT:	utotal += total;

/*********************************************************************
*	If the backup data is being saved on disk, write an 
*	END_OF_RECORD_MARKER to separate the data sets. 
**********************************************************************/
		if(!error){
		  if (backup and WriteDEFiles(NULL, END_OF_RECORD_MARKER, 0, 0))	/* end of file index record */
			break;
		}

/*********************************************************************
*	Scan for the next data set - if another data set is available,
*	the loop continues; otherwise display the number of resources
*	scanned and terminate.
**********************************************************************/
		if ((ccode = NWSMTSScanNextDataSet(connection, &sequence, &scanInfo, &dataSetNameList)) isnt 0
				and ccode isnt NWSMTS_NO_MORE_DATA_SETS)
			PrintError(ccode, "NWSMTSScanNextDataSet", 0);
	}

	printf("\n\aNumber of resources scanned: %u\n", resourceCount);

/*********************************************************************
*	This is the common termination point for the backup function.
*	If the restart flag has not been set to FALSE, go back to loop
*	and start over.
*	If we were doing a backup (not just scan) display some statistics.
**********************************************************************/
Terminate:
	if (restart)
	{
		if (command is 'B')
		{
			float kb = utotal / 1024.0, min = ElapsedTime() / 60000.0;

			printf("Total read:   %.2f KB\n", kb);
			printf("Elapsed time: %.2f minutes\n", min);
			printf("Throughput:   %.2f MB/min\n", kb / min / 1024.0);
			pause();
		}

		goto Loop;
	}

/*********************************************************************
*	That's it for the Backup/Scan - do some clean up and return.
**********************************************************************/
	if (backup)
		CloseDEFiles();

	if (resourceName)
		free(resourceName);
	if(saveFirstResourceList)
		NWSMFreeNameList(&saveFirstResourceList);
}
/****************************************************************************/

/*********************************************************************
* NAME: Restore
*
* WHAT: This function controls the overall process of restoring data
*	to a Target Service. A complimentary routine (Backup) exists
*	for backing up Target Service data.
*
* SMS-SPECIFIC ROUTINES CALLED: NWSMTSSetRestoreOptions (TSAPI)
*				NWSMTSCloseDataSet (TSAPI)
*				NWSMTSOpenDataSetForRestore (TSAPI)
*				NWSMTSWriteDataSet (TSAPI)
*
**********************************************************************/
void Restore(void)
{
	BUFFER *SMSData;
	CCODE ccode;
	int command, restart = TRUE;

	UINT8 overwriteParent = TRUE, overwriteChild = TRUE, parentFlag, terminalNodeNameOnlyFlag;
	UINT16 resourceCount;
	UINT32 bytesToWrite, dataSetHandle = 0, parentHandle = 0, total, utotal;
	UINT8 parentWasExcluded = FALSE;


Loop:
	total = utotal = 0;
	resourceCount = 0;
	ccode = 0;

/*********************************************************************
*	Prompt the user to select the option(s) of overwriting/not overwriting
*	the Parent/Child data set.
**********************************************************************/
	switch (command = GetRCommand(overwriteParent, overwriteChild))
	{
	case 'R':
		break;

	case 'P':
		overwriteParent = !overwriteParent;
		goto Loop;

	case 'C':
		overwriteChild = !overwriteChild;
		goto Loop;

	default:
		restart = FALSE;
		goto Terminate;
	}

/*********************************************************************
*	If a selection list exists - set up the restore options so
*	that data sets will be checked against the list and the appropriate
*	data sets excluded.
**********************************************************************/
	if (selectionList)
		NWSMTSSetRestoreOptions(connection, scanControl.generateCRC, FALSE, selectionList);

	clrscr();

/*********************************************************************
*	For this example engine, backed up data is stored in DOS files
*	on disk instead of tape (or other backup media). Here we open
*	the disk files for reading. Two files are maintained - an index
*	file and a data file.
**********************************************************************/
	OpenDEFiles('r');

	StartTimer();


	while (!ccode)
	{
		UINT16 mode;

/*********************************************************************
* 	Read the backup data from the DOS files 
**********************************************************************/
/*********************************************************************
*	NOTE: The first entry in the backup data index file is an
*	END_OF_RECORD_MARKER - so the first read should return this
*	marker.
**********************************************************************/
		if ((SMSData = ReadDEFiles(&bytesToWrite, &parentFlag, &terminalNodeNameOnlyFlag)) is NULL)
			break;

/*********************************************************************
*	If we have reached an END_OF_RECORD_MARKER in the index file
*	on the disk - increment the "total bytes restored" count and
*	resource count. Then close the data set if it is open (handle
*	is non-NULL).
*	If not END_OF_RECORD_MARKER then update the byte count and
*	attempt to write the data to the Target Service.
**********************************************************************/
		if (bytesToWrite is END_OF_RECORD_MARKER)	/* end of file */
		{
			utotal += total;
			total = 0;
			resourceCount++;
			if (dataSetHandle)
			{
				status(" Close");
				NWSMTSCloseDataSet(connection, &dataSetHandle);
				dataSetHandle = NULL;
			}

/*********************************************************************
*	Do another read of the backup data on disk to determine if
*	another data set follows the END_OF_RECORD_MARKER. If the read
*	returns a NULL then we have come to the end of the backup data.
*	If the read returns a non-NULL pointer - attempt to open the data
*	set with the appropriate parameters as indicated in the following
*	truth table:
*
*	ParentFlag  TerminalNodeNameFlag  |  OpenDataSetForRestore call
*	---------------------------------------------------------------
*	FALSE	    FALSE		  |  Don't use Parent Handle
*	FALSE	    TRUE		  |  USE Parent Handle
*	TRUE	    FALSE		  |  Don't use Parent Handle
*	TRUE	    TRUE		  |  SET Parent Handle
**********************************************************************/
			if ((SMSData = ReadDEFiles(&bytesToWrite, &parentFlag, &terminalNodeNameOnlyFlag)) isnt NULL)
			{
				/* We don't want to open any terminal node
				 * children of a parent that has been 
				 * excluded */
				if(parentFlag) parentWasExcluded=FALSE;
				if(parentWasExcluded) continue;

				if (parentFlag and terminalNodeNameOnlyFlag)
				{
					if (parentHandle)
					{
						status("\nClose parent\n");
						NWSMTSCloseDataSet(connection, &parentHandle);
						parentHandle = NULL;
					}

					mode = overwriteParent ? NWSM_OVERWRITE_DATA_SET : NWSM_CREATE_PARENT_HANDLE;
					status("\nCreate parent");
					if ((ccode = NWSMTSOpenDataSetForRestore(connection, 0L, NULL /* restoreDataSetName */,
							mode, &parentHandle)) isnt 0)
						PrintError(ccode, "NWSMTSOpenDataSetForRestore", __LINE__);
				}
	
				else if (terminalNodeNameOnlyFlag)
				{
					mode = overwriteChild ? NWSM_OVERWRITE_DATA_SET : NWSM_DO_NOT_OVERWRITE_DATA_SET;
					status("\nOpen child /w parent handle");
					if ((ccode = NWSMTSOpenDataSetForRestore(connection, parentHandle, NULL, mode, &dataSetHandle)) isnt 0)
						PrintError(ccode, "NWSMTSOpenDataSetForRestore", __LINE__);
				}

				else
				{
					mode = overwriteChild ? NWSM_OVERWRITE_DATA_SET : NWSM_DO_NOT_OVERWRITE_DATA_SET;
					status(parentFlag ? "\nOpen parent, parent handle not used" : "\nOpen child w/o parent handle");
					if ((ccode = NWSMTSOpenDataSetForRestore(connection, 0L, NULL, mode, &dataSetHandle)) isnt 0)
						PrintError(ccode, "NWSMTSOpenDataSetForRestore", __LINE__);
				}
			}
	
			else		/* end of backup data */
				break;
		}

/***************************************************************************
*	Increment the byte count and display the count to the user.
***************************************************************************/
		/* We don't want to write any terminal node children of
		 * a parent that has been excluded */
		if(parentWasExcluded) continue;

		total += bytesToWrite;
		gotoxy(60, wherey());
		status(" %-8lu ", total);

/*********************************************************************
*	Attempt to write the data set to the Target Service. If 
*	successful - continue reading more backup data.
*	If the write fails - display the reason, skip this data set
*	and look for another data set to restore.
**********************************************************************/
		if ((ccode = NWSMTSWriteDataSet(connection, parentFlag and
				terminalNodeNameOnlyFlag ? parentHandle : dataSetHandle, bytesToWrite, SMSData)) isnt 0)
		{
			if (ccode is NWSMTS_VALID_PARENT_HANDLE)
			{
				status(" Valid");
				total = 0;
				ccode = 0;
			}

			else if (ccode is NWSMTS_DATA_SET_EXCLUDED)
			{
				gotoxy(60, wherey());
				status(" Excluded ");
				total = 0;
				ccode = 0;
				if(parentFlag) parentWasExcluded = TRUE;
			}

			else if (ccode is NWSMTS_DATA_SET_ALREADY_EXISTS)
			{
				gotoxy(60, wherey());
				status(" Exists   ");
				ccode = 0;
				total = 0;
			}

			else 
				PrintError(ccode, "NWSMTSWriteDataSet", 0);

/*********************************************************************
*	Flush this data set -- Data Sets are separated by 
*	END_OF_RECORD_MARKERs
**********************************************************************/

			do
				ReadDEFiles(&bytesToWrite, &parentFlag, &terminalNodeNameOnlyFlag);
			while (bytesToWrite isnt END_OF_RECORD_MARKER);

/*********************************************************************
*	The loop immediately above will read up to and including
*	an END_OF_RECORD_MARKER -- This call to UngetEOF prevents
*	the next ReadDEFiles call from reading any more data until
*	the END_OF_RECORD_MARKER has been processed.
**********************************************************************/
			UngetEOF();
		}
	}

/*********************************************************************
*	This is the common termination point for the Restore function.
*	Do some clean up then check the restart flag. If restart hasn't
*	been set to false - display some statistics (if this was a restore)
*	then go back to loop and start over.
**********************************************************************/
Terminate:
	if (dataSetHandle)
	{
		status(" Close\n");
		NWSMTSCloseDataSet(connection, &dataSetHandle);
		dataSetHandle = 0;
	}

	if (parentHandle)
	{
		status("\nClose parent\n");
		NWSMTSCloseDataSet(connection, &parentHandle);
		parentHandle = 0;
	}

	CloseDEFiles();

	if (restart)
	{
		if (command is 'R')
		{

			float kb = utotal / 1024.0, min = ElapsedTime() / 60000.0;
		/*	The index file begins with a END_OF_RECORD_MARKER so we need to decrement resourceCount */
			printf("\n\n\aNumber of resources restored: %u\n", --resourceCount);
			printf("Total written: %.2f KB\n", kb);
			printf("Elapsed time:  %.2f minutes\n", min);
			printf("Throughput:    %.2f MB/min\n", kb / min / 1024.0);
			pause();
		}

		goto Loop;
	}

	if (selectionList)
		free(selectionList);
}
/****************************************************************************/

/*********************************************************************
* NAME: EraseFile
*
* WHAT: Remove a selected Data Set
*
* SMS-SPECIFIC ROUTINES CALLED: NWSMPutOneName (SMS UTILITY API)
*				NWSMTSScanDataSetBegin (TSAPI)
*				NWSMTSScanNextDataSet (TSAPI)
*				NWSMGetDataSetName (SMS UTILITY API)
*				NWSMTSDeleteDataSet (TSAPI)
*				NWSMTSScanDataSetEnd (TSAPI)
*
**********************************************************************/
void EraseFile(void)
{
	CCODE ccode;
	UINT32 sequence = 0L;
	char myName[30], mode;

	NWSM_DATA_SET_NAME dataSetName;
	NWSM_DATA_SET_NAME_LIST *resourceName = NULL, *dataSetNameList = NULL;

	scanInfo = NULL;

/*********************************************************************
*	Prompt the user to select whether to delete a file or a
*	directory. Then get the name of the file/directory to delete.
**********************************************************************/
	printf("\nDelete File or Directory? (f/d): "), mode = toupper(getch());
	printf("\nEnter %s to %s: ", mode is 'F' ? "file" : "directory", "delete: ");
	gets(myName), strupr(myName);

/*********************************************************************
*	Create a data set name list with an entry for the above selected
*	file/directory.
**********************************************************************/
	if (myName[strlen(myName) - 1] is ':')
			ccode = NWSMPutOneName((void HUGE **)&resourceName, NWSM_TSA_DEFINED_RESOURCE_TYPE, 0L, 0,
				NULL, NULL, myName);
	else
			ccode = NWSMPutOneName((void HUGE **)&resourceName, DOSNameSpace, 0L, 0,":", "/", myName);

	if (ccode)
	{
		PrintError(ccode, "NWSMPutOneName", 0);
		goto Terminate;
	}

/*********************************************************************
*	Scan for the selected file/directory. If a directory was selected,
*	then the NWSMTSScanDataSetBegin will be sufficient to return the
*	correct data set. If a file was selected (mode = F), another
*	scan is needed since the first scan gets the directory (parent)
*	and the second scan returns the desired file (child).
**********************************************************************/
	if ((ccode = NWSMTSScanDataSetBegin(connection, resourceName, &scanControl, selectionList, &sequence, &scanInfo, &dataSetNameList)) isnt 0)
	{
		PrintError(ccode, "NWSMTSScanDataSetBegin", 0);
		goto Terminate;
	}

	if (mode is 'F')
	{
		if ((ccode = NWSMTSScanNextDataSet(connection, &sequence, &scanInfo, &dataSetNameList)) isnt 0)
		{
			PrintError(ccode, "NWSMTSScanNextDataSet", __LINE__);
			goto Terminate;
		}
	}

/*********************************************************************
*	This call to NWSMGetDataSetName provides the name that results
*	from the above scan. This is only used to provide feedback to 
*	the user.
**********************************************************************/
	NWSMGetDataSetName(dataSetNameList, 0L, &dataSetName);
	printf("Delete: %s\n", dataSetName.name);

/*********************************************************************
* Delete the selected file/directory 
**********************************************************************/
	if ((ccode = NWSMTSDeleteDataSet(connection, sequence)) isnt 0)
		PrintError(ccode, "NWSMTSDeleteDataSet", 0);

/*********************************************************************
* Close the scan and clean up 
**********************************************************************/
Terminate:
	NWSMTSScanDataSetEnd(connection, &sequence, &scanInfo, &dataSetNameList);

	if (resourceName)
		free(resourceName);

	pause();
	return;
}
/****************************************************************************/

/*********************************************************************
* NAME: RenameFile
*
* WHAT: Rename a selected data set.
*
* SMS-SPECIFIC ROUTINES CALLED: NWSMPutOneName (SMS UTILITY API)
*				NWSMTSScanDataSetBegin (TSAPI)
*				NWSMTSScanNextDataSet (TSAPI)
*				NWSMGetDataSetName (SMS UTILITY API)
*				NWSMTSRenameDataSet (TSAPI)
*				NWSMTSScanDataSetEnd (TSAPI)
*
**********************************************************************/
void RenameFile(void)
{
	CCODE ccode;
	UINT32 sequence = 0L;
	char myName[40], newName[20], mode;

	NWSM_DATA_SET_NAME dataSetName;
	NWSM_DATA_SET_NAME_LIST *resourceName = NULL, *dataSetNameList = NULL;

	scanInfo = NULL;

/*********************************************************************
*	Prompt the user to select whether to rename a file or a directory.
*	Then get the name of the file/directory to rename.
**********************************************************************/
	printf("\nRename File or Directory? (f/d): "), mode = toupper(getch());
	printf("\nEnter %s to %s: ", mode is 'F' ? "file" : "directory", "rename"), gets(myName);

/*********************************************************************
*	Create a data set name list with an entry for the above
*	selected file/directory.
**********************************************************************/
	if (myName[strlen(myName) - 1] is ':')
			ccode = NWSMPutOneName((void HUGE **)&resourceName, NWSM_TSA_DEFINED_RESOURCE_TYPE, 0L, 0,
				NULL, NULL, myName);
	else
			ccode = NWSMPutOneName((void HUGE **)&resourceName, DOSNameSpace, 0L, 0,":", "/", myName);

	if (ccode)
	{
		PrintError(ccode, "NWSMPutOneName", 0);
		goto Terminate;
	}

/*********************************************************************
*	Scan for the directory that was selected above or for the
*	parent directory if a file was selected for renaming.
**********************************************************************/
	if ((ccode = NWSMTSScanDataSetBegin(connection, resourceName, &scanControl, selectionList, &sequence, &scanInfo, &dataSetNameList)) isnt 0)
	{
		PrintError(ccode, "NWSMTSScanDataSetBegin", 0);
		goto Terminate;
	}

/*********************************************************************
*	Prompt for the new name of the file/directory.
**********************************************************************/
	printf("Enter new %s %s: ", mode is 'F' ? "file" : "directory", "name"), gets(newName), strupr(newName);

/*********************************************************************
*	If a file was selected above (mode = F) do another scan since
*	the first scan only returned the parent of this file.
**********************************************************************/
	if (mode is 'F')
	{
		if ((ccode = NWSMTSScanNextDataSet(connection, &sequence, &scanInfo, &dataSetNameList)) isnt 0)
		{
			PrintError(ccode, "NWSMTSScanNextDataSet", __LINE__);
			goto Terminate;
		}
	}

/*********************************************************************
*	This call to NWSMGetDataSetName provides the name that resulted
*	from the above scan. This is only used to provide feedback to
*	the user.
**********************************************************************/
	NWSMGetDataSetName(dataSetNameList, 0L, &dataSetName);

/*********************************************************************
*	Display the name of the file/directory returned by the scan.
**********************************************************************/
	printf("Rename %s to %s\n", dataSetName.name, newName);

/*********************************************************************
* Rename the selected file 
**********************************************************************/
	if ((ccode = NWSMTSRenameDataSet(connection, sequence, DOSNameSpace, newName)) isnt 0)
		PrintError(ccode, "NWSMTSRenameDataSet", 0);

/*********************************************************************
* Close the scan and clean up 
**********************************************************************/
Terminate:
	NWSMTSScanDataSetEnd(connection, &sequence, &scanInfo, &dataSetNameList);
	if (resourceName)
		free(resourceName);

	pause();
	return;
}
/****************************************************************************/


char *buildDateAndTime, moduleName[MAXFILE], *moduleVersion;

/*********************************************************************
* NAME: main
*
* WHAT: This is the main function for the demonstration engine.
*	List the TSAs that are visible on the network. Prompt the
*	user to select a TSA then attempt to establish a connection
*	with that TSA.
*	After connecting to a TSA enter an infinite loop in which user
*	commands are received and processed.
*
* SMS-SPECIFIC ROUTINES CALLED: 
*				EstablishConnection (DE1.C)
*				ListTSs (DE1.C)
*				ListScanTypeStrings (DE1.C)
*				ListSelectionStrings (DE1.C)
*				BuildSelectionList (DE1.C)
*				DisplaySelectionList (DE1.C)
*				FreeSelectionList (DE1.C)
*				EraseFile (DE1.C)
*				RenameFile (DE1.C)
*				Backup (DE1.C)
*				Restore (DE1.C)
*				NWSMTSReleaseTargetService (TSAPI)
*				NWSMReleaseTSA (TSAPI)
*
**********************************************************************/
void main(int argc, char *argv[])
{

	char drive[MAXDRIVE], dir[MAXDIR], ext[MAXEXT];

	argc = argc;
	fnsplit(argv[0], drive, dir, moduleName, ext);
	buildDateAndTime = __DATE__ " @ " __TIME__;
	moduleVersion = "v3.11";


/***FOR NLM CONVERSION***
* Since the Novell/Watcom compiler/linker is needed for generating
* NLMs and since the "fnsplit" function is not in the Novell/Watcom
* libraries, replace the above call to fnsplit with the following:
*
* 	moduleName = "NE" <--- replace NE with the name of your NLM.
*
* Also remove the following:
*				The arguments passed to main argc and argv.
*				The definitions of drive,dir and ext.
*				The line argc = argc;
*
* Redefine moduleName[MAXFILE] as char *moduleName.
*
* Remove the following call to NWCallsInit().
**********************************************************************/

	setcbrk(1);
	NWCallsInit(NULL, NULL); /*need to call this before any library calls*/

	if (EstablishConnection())
	{
		puts("Connection denied.");
		exit(0);
	}

	loop
	{
		switch (GetCommand())
		{

		case 'C':
			BuildSelectionList();
			DisplaySelectionList();
			break;

		case 'D':
			DisplaySelectionList();
			break;

		case 'F':
			FreeSelectionList();
			break;

		case 'E':
			EraseFile();
			break;

		case 'N':
			RenameFile();
			break;

		case 'B':
			Backup();
            /* before releasing connection to TSA, get "ERROR LOG" and "SKIPPED DATA SET" */
            ProcessErrorLog();
            pause();
            ProcessSkippedDataSets();
            pause();
			break;

		case 'R':
			Restore();
			break;

		case 'Q':
			if (connection)
			{
				NWSMTSReleaseTargetService(&connection);
				NWSMReleaseTSA(&connection);
			}
		
			if (selectionList)
				free(selectionList);

/***FOR NLM CONVERSION***/
/* Be sure to clean up and release all resources before calling exit
* if you have added any code of your own to this file--otherwise a nasty
* message is displayed on the server console.
*************************************************************************/
			exit(0);
		}
	}
}

/****************************************************************************/
/*********************************************************************
* NAME: ProcessErrorLog
*
* WHAT: Fetchs and displays "ERROR LOG" data set from the Target Service.
*
* SMS-SPECIFIC ROUTINES CALLED: 
*				NWSMPutOneName (SMS UTILITY API)
*				NWSMTSScanDataSetBegin (TSAPI)
*				NWSMTSOpenDataSetForBackup (TSAPI)
*				NWSMTSReadDataSet (TSAPI)
*				NWSMTSCloseDataSet (TSAPI)
*				NWSMTSScanDataSetEnd (TSAPI)
*
**********************************************************************/
void ProcessErrorLog(void)
{
  /* connection is global value in this example */
  /* dataBuffer is global value in this example */
  CCODE                      ccode;
  NWSM_DATA_SET_NAME_LIST   *resourceName = NULL;
  NWSM_DATA_SET_NAME_LIST   *dataSetNameList = NULL;
  UINT32                     sequence;
  UINT32                     mode = NWSM_USE_LOCK_MODE_IF_DW_FAILS;
  UINT32                     bytesRead;
  UINT32                     dataHandle = NULL;
  UINT16                     readSize = BUFFER_SIZE;
  int                        empty = TRUE;

  /* display error log name */
  printf ("\n\n\"ERROR LOG\":\n");

  /* setup resource name list */
  if ((ccode = NWSMPutOneName( 
                          &resourceName, 
                          NWSM_TSA_DEFINED_RESOURCE_TYPE,
                          0L,                              /* selectionType */
                          FALSE,                           /* reverse order */
                          NULL,                          /* first separator */
                          NULL,                         /* second separator */
                          "ERROR LOG" ))          /* Error log special name */
       != 0 )
  {
    PrintError( 
      ccode, 
      "DE: * * *   ERROR  * * * NWSMPutOneName(ERROR LOG)", 
      0);
    goto Terminate;
  }

  /* read in error log data set */
  if ((ccode = NWSMTSScanDataSetBegin( 
                                 connection,
                                 resourceName,
                                 NULL,                      /* scan control */
                                 NULL,                    /* selection list */
                                 &sequence,
                                 NULL,                  /* scan information */
                                 &dataSetNameList ))
       != 0 )
  {
    PrintError( 
      ccode, 
      "DE: * * *   ERROR   * * * NWSMTSScanDataSetBegin(ERROR LOG)", 
      0);
    goto Terminate;
  }
  if (((ccode = NWSMTSOpenDataSetForBackup( 
                                          connection,
                                          sequence,
                                          0L,                       /* mode */
                                          &dataHandle ))
       !=0 ) || ( dataHandle == 0L ) )
  {
    PrintError( 
      ccode, 
      "DE: * * *   ERROR   * * * NWSMTSOpenDataSetForBackup(ERROR LOG)", 
      0);
    goto Terminate;
  }
  do
  {
    if (( ccode = NWSMTSReadDataSet( 
                               connection,
                               dataHandle,
                               readSize,
                               &bytesRead,
                               dataBuffer ))
          != 0 )
    {
      PrintError( 
        ccode, 
        "DE: * * *   ERROR   * * * NWSMTSReadDataSet(ERROR LOG)", 
        0);
      break;
    }
    if (bytesRead)
    {
      dataBuffer[ bytesRead ] = 0;
      printf( dataBuffer );
      empty = FALSE;
    }
    if ( bytesRead < readSize )
    {
      break;
    }
  } while ( ccode == 0 );
Terminate:

  if (empty)
  {
    printf( "\t<empty>\n");
  }

  if ( dataHandle )
  {
    if (( ccode = NWSMTSCloseDataSet(
                                connection,
                                &dataHandle ))
          != 0 && ccode != NWSMTS_INVALID_DATA_SET_HANDLE )
    {
      PrintError( 
        ccode, 
        "DE: * * *   ERROR   * * * NWSMTSCloseDataSet(ERROR LOG)", 
        0);
    }
  }
  if (( ccode = NWSMTSScanDataSetEnd( 
                                connection,
                                &sequence,
                                NULL, /* scan information */
                                &dataSetNameList ))
        != 0 )
  {
    PrintError( 
      ccode, 
      "DE: * * *   ERROR   * * * NWSMTSScanDataSetEnd(ERROR LOG)", 
      0);
  }
  if ( resourceName )
    free( resourceName );
  if ( dataSetNameList )
    free( dataSetNameList );

} /* ProcessErrorLog() */

/****************************************************************************/
/*********************************************************************
* NAME: ProcessSkippedDataSets
*
* WHAT: Fetchs and displays "SKIPPED DATA SETS" data set from the 
*       Target Service.
*
* SMS-SPECIFIC ROUTINES CALLED: 
*				NWSMPutOneName (SMS UTILITY API)
*				NWSMTSScanDataSetBegin (TSAPI)
*				NWSMTSOpenDataSetForBackup (TSAPI)
*				NWSMTSReadDataSet (TSAPI)
*				NWSMTSCloseDataSet (TSAPI)
*				NWSMTSScanDataSetEnd (TSAPI)
*
**********************************************************************/
void ProcessSkippedDataSets(void)
{
  /* connection is global value in this example */
  /* dataBuffer is global value in this example */
  CCODE                      ccode;
  NWSM_DATA_SET_NAME_LIST   *resourceName = NULL;
  NWSM_DATA_SET_NAME_LIST   *dataSetNameList = NULL;
  NWSM_DATA_SET_NAME         dataSetName;
  UINT32                     sequence;
  UINT32                     mode = NWSM_USE_LOCK_MODE_IF_DW_FAILS;
  UINT32                     bytesRead;
  UINT32                     dataHandle = NULL;
  UINT32                     bytesToRead;
  UINT32                     leftOverBytes = 0L;
  UINT16                     dataSetNameSize;
  BUFFER                    *bufferEnd = NULL;
  BUFFER                    *bufferPtr = NULL;
  BUFFER                    *readPtr = NULL;
  int                        empty = TRUE;
  int                        getNextName;
  char                       errorMessage[256];

  /* display SKIPPED DATA SETS name */
  printf ("\n\n\"SKIPPED DATA SETS\":\n");

  /* setup resource name list */
  if ((ccode = NWSMPutOneName( (void HUGE **)
                          &resourceName, 
                          NWSM_TSA_DEFINED_RESOURCE_TYPE,
                          0L,                              /* selectionType */
                          FALSE,                           /* reverse order */
                          NULL,                          /* first separator */
                          NULL,                         /* second separator */
                          (void HUGE *)
                          "SKIPPED DATA SETS" )) /* SKIPPED DATA SETS special name */
       != 0 )
  {
    PrintError( 
      ccode, 
      "DE: * * *   ERROR  * * * NWSMPutOneName(SKIPPED DATA SETS)", 
      0);
    goto Terminate;
  }

  /* read in SKIPPED DATA SETS data set */
  if ((ccode = NWSMTSScanDataSetBegin( 
                                 connection,
                                 resourceName,
                                 NULL,                      /* scan control */
                                 NULL,                    /* selection list */
                                 &sequence,
                                 NULL,                  /* scan information */
                                 &dataSetNameList ))
       != 0 )
  {
    PrintError( 
      ccode, 
      "DE: * * *   ERROR   * * * NWSMTSScanDataSetBegin(SKIPPED DATA SETS)", 
      0);
    goto Terminate;
  }
  if (((ccode = NWSMTSOpenDataSetForBackup( 
                                          connection,
                                          sequence,
                                          0L,                       /* mode */
                                          &dataHandle ))
       !=0 ) || ( dataHandle == 0L ) )
  {
    PrintError( 
      ccode, 
      "DE: * * *   ERROR   * * * NWSMTSOpenDataSetForBackup(SKIPPED DATA SETS)", 
      0);
    goto Terminate;
  }

  /* SKIPPED DATA SETS data set is a binary data set layed out in
   * NWSM_DATA_SET_NAME_LIST structure format. Leftover pieces may 
   * be in read buffer. Need to manage the buffer to handle this.
   */

  readPtr = dataBuffer; /* 1st time through, start at beginning */
  leftOverBytes = 0L;
  bytesRead = 0L;
  empty = TRUE;
  do
  {
    bytesToRead = BUFFER_SIZE - leftOverBytes;
    if (( ccode = NWSMTSReadDataSet( 
                               connection,
                               dataHandle,
                               bytesToRead,
                               &bytesRead,
                               readPtr ))
          == 0 )
    {
      if ( ! bytesRead )
      {
        break;
      }
      /* lets parse buffer, output name(s) and check boundary conditions
       * of buffer
       */
      bufferPtr = dataBuffer;
      bufferEnd = dataBuffer + bytesRead;
      getNextName = TRUE;
      while ( getNextName )
      {
        /* make sure there's enough space in buffer to contain a whole 
         * data set. If not, shuffle buffer */
        if (( leftOverBytes = bufferEnd - bufferPtr ) < 4 )
        {
          getNextName = FALSE;
        }
        else
        {
          /* data set name size is within first 4 bytes (2nd UINT16):
           * if data set name size == left over bytes, could handle, but
           * an increment off end will occur, so treat the same as 
           * greater-than condition.
           */
          if (( dataSetNameSize = ((NWSM_DATA_SET_NAME_LIST *) bufferPtr )->
                             dataSetNameSize )
                >= leftOverBytes )
            getNextName = FALSE;
        }
        if (! getNextName )
        {
          /* move left over to top of buffer */
          memcpy( dataBuffer, bufferPtr, leftOverBytes );

          /* prepare for next read */
          readPtr = dataBuffer + leftOverBytes;
          break;
        }
        if ((ccode = NWSMGetOneName( bufferPtr, &dataSetName )) == 0 )
        {
          NWSMConvertError( 
                              connection, 
                              dataSetName.selectionType,
                              errorMessage );
          printf ("\t(0x%08lX) %s: %s\n", 
                     dataSetName.selectionType, 
                     errorMessage, 
                     dataSetName.name );
        }
        else
        {
          PrintError( 
            ccode, 
            "DE: * * *   ERROR   * * * NWSMGetOneName(SKIPPED DATA SETS)", 
            0);
          empty = FALSE;
          break;
        }
        bufferPtr += dataSetNameSize + 2;
      } /* end while ( getNextName ) */
      empty = FALSE;
    } /* end read data set OK */
    else
    {
      PrintError( 
        ccode, 
        "DE: NWSMTSReadDataSet(SKIPPED DATA SETS)", 
        0);
      break;
    }
  } while ( 1 );

Terminate:

  if (empty)
  {
    printf( "\t<empty>\n");
  }

  if ( dataHandle )
  {
    if (( ccode = NWSMTSCloseDataSet(
                                connection,
                                &dataHandle ))
          != 0 && ccode != NWSMTS_INVALID_DATA_SET_HANDLE )
    {
      PrintError( 
        ccode, 
        "DE: * * *   ERROR   * * * NWSMTSCloseDataSet(SKIPPED DATA SETS)", 
        0);
    }
  }
  if (( ccode = NWSMTSScanDataSetEnd( 
                                connection,
                                &sequence,
                                NULL, /* scan information */
                                &dataSetNameList ))
        != 0 )
  {
    PrintError( 
      ccode, 
      "DE: * * *   ERROR   * * * NWSMTSScanDataSetEnd(SKIPPED DATA SETS)", 
      0);
  }
  if ( resourceName )
    free( resourceName );
  if ( dataSetNameList )
    free( dataSetNameList );

} /* ProcessSkippedDataSets() */

/****************************************************************************/
/****************************************************************************/

