/*****************************************************************************
 * Program Name:  Storage Management Services (Example Engine)
 *
 * Filename:	  de2.c
 *
 * Date Created:  March 14, 1991
 *
 * Version:		  1.0
 *
 * Files used:	  de.h
 *
 * Functions Defined In This File:  DisplayList	
 *				  getun
 *				  getpw
 *				  pause
 *				  printvi
 *				  StartTimer
 *				  ElapsedTime
 *				  GetCommand
 *				  GetBCommand
 *				  GetRCommand
 *				  GetNameSpaceType
 *				  GetSelectionType
 *				  SetNameSpaceType
 *				  SetSelectionType
 *				  WriteDEFiles
 *				  ReadDEFiles
 *				  OpenDEFiles
 *				  CloseDEFiles
 *
 * Date Modified: 
 *
 * Modifications: 
 *
 *
 * $Lgb$
1.1 03/18/92 Gary For SMS v1.1 SDK
1.2 04/23/92 Gary For SMS v1.2 SDK (kit is already released
1.3 04/29/92 Gary no change from 1.2
1.4 07/27/92 tony Updated for use with SMDR
 * $Lge$
 *
 *
 *
 * Comments:	    DE is a sample SME
 *
 * (C) Unpublished Copyright of Novell, Inc.  All Rights Reserved.
 *
 * No part of this file may be duplicated, revised, translated, localized or
 * modified in any manner or compiled, linked or uploaded or downloaded to or
 * from any computer system without the prior written consent of Novell, Inc.
 ****************************************************************************/

 /***For NLM Conversion***/
 /****************************************************************************
 *	NOTE: Comments are included with this code that describe how to
 *		convert this demonstration engine to an NLM. To make an NLM 
 *		well behaved, control should be passed to other processes on
 *		the server at least every 150 milliseconds (386 @ 25Mhz).
 *		To accomodate this, calls to the ThreadSwitch function (CLIB) 
 *		should be appropriately scattered through the code. 
 *		ThreadSwitches should be considered for any sections of
 *		code, such as loops, which could execute for more 
 *		than 150 milliseconds. CLIB routines that block should already
 *		pass control to other threads so you won't have to worry about
 *		these.
 *
 *		To find the changes need for converting this demonstration 
 *		to an NLM, search for the string "NLM" in this code.
 ****************************************************************************/

#include <sms.h>
#include "de.h"
extern UINT32 connection;


/*********************************************************************
* NAME: DisplayList
*
* WHAT: Display the names (strings) in a linked list that is passed
*	as a parameter to this function. Specifically a list structure
*	of type NWSM_NAME_LIST is displayed by this function.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
CCODE DisplayList(NWSM_NAME_LIST *list)
{
	CCODE ccode = 0;
	NWSM_NAME_LIST *ptr, *next;
    int count = 0;

	if (!list)
		goto Return;

	ptr = list;


	while (ptr)
	{
		next = ptr->next;
		printf("%-16s", ptr->name);
		printf("\n");
		ptr = next;
        ++count;
/*        if ( ! (count % 20) )
          pause();
*/
	}
	printf("\n\n");

Return:
	return (ccode);
}
/****************************************************************************/

/*********************************************************************
* NAME: getun
*
* WHAT: Prompt the user to enter a user name. The name "ADMIN"
*	is used as a default.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
void getun(char *user)
{
	char firstChar;
	int x, y;

	y = wherey(), x = wherex();
	printf("ADMIN");
	gotoxy(x, y);
	if ((firstChar = getch()) is '\r')
	{
		strcpy(user, "ADMIN");
		puts("");
	}

	else
	{
		puts("          ");
		gotoxy(x, y);
		putch(firstChar);
		*user = firstChar;
		gets(user + 1);
		strupr(user);
	}
}
/****************************************************************************/

/*********************************************************************
* NAME: getpw
*
* WHAT: Get a password from the user to be used in conjunction with
*	the user name.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
void getpw(char *password)
{
	int count = 2;

	while (count)
	{
		switch (password[count] = toupper(getch()))
		{
		case '\r':
			*(UINT16 *)password = count - 2;
			count = 0;
			break;

		case '\b':
			if (count > 2)
				--count;

			else
				putch('\a');
			break;

		default:
			++count;
			break;
		}
	}
}
/****************************************************************************/

/*********************************************************************
* NAME: pause
*
* WHAT: Wait until the user enters a keystroke.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
void pause(void)
{

	while (kbhit()) getch();
	puts("Strike any key to continue..."), getch();
	while (kbhit()) getch();
	clrscr();
}
/****************************************************************************/


/***FOR NLM CONVERSION***
* 	Remove this declaration of buf
************************/
struct timeb buf;

/*********************************************************************
* NAME: StartTimer
*
* WHAT: Start the timer used by the ElapsedTime function.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
void StartTimer(void)
{
	ftime(&buf);
}

/*********************************************************************
* NAME: ElapsedTime
*
* WHAT: Return the elapsed time (in milliseconds) since the StartTimer
*	function was last called.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
UINT32 ElapsedTime(void)
{
	struct timeb now;

	ftime(&now);
	return (now.time * 1000 + now.millitm) - (buf.time * 1000 + buf.millitm);
}
/****************************************************************************/

/***FOR NLM CONVERSION***
* Since the Novell/Watcom compiler/linker is needed for generating
* NLMs and since the "ftime" function is not in the Novell/Watcom
* libraries, replace the StartTimer and ElapsedTime functions above with
* code similar to the following:
*	clock_t begin_time;
*
*	void StartTimer(void)
*	{
*		begin_time = clock();
*	}
*
*	UINT32 ElapsedTime(void)
*	{
*	clock_t time_now;
*	
*		time_now = clock();
*		return((UINT32)((time_now - begin_time)*10) );
*	
*	
*	}
*
*	NOTE: This is just an example of how to replace ftime(), the timer
*	granularity is not as good as ftime() -- hundredths of seconds instead
*	of milliseconds.
**********************************************************************/

/*********************************************************************
* NAME: GetCommand
*
* WHAT: Prompt the user to enter a command, then get the command
*	and return it to the caller. The valid commands are:
*		Create Selection List
*		Display Selection List
*		Free Selection List
*		Erase File
*		Rename File
*		List TSA strings
*		Backup
*		Restore
*		Quit
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
char GetCommand(void)
{
	char command;
	int x, y;

	clrscr();
	printf(	"C: Create SelectionList\n"
			"D: Display SelectionList\n"
			"F: Free SelectionList\n"
			"E: Erase (Delete) File\n"
			"N: New File Name (Rename File)\n"
			"B: Backup\n"
			"R: Restore\n\n"
			"Q: Quit\n" 
			"\nSelect option: ");

	y = wherey(), x = wherex();


	loop
	{
		gotoxy(x, y);
		printf(" ");
		gotoxy(x, y);

		command = toupper(getch());
		if (strchr("CDFENLVBRQ", command))
			goto Return;

		putch('\a');
	}

Return:
	return (command);
}
/****************************************************************************/

/*********************************************************************
* NAME: GetBCommand
*
* WHAT: Prompt the user to select one of the following Backup/Scan options:
*
*	Backup Resource
*	Backup Subdirectory
*	Scan Resource
*	Scan Subdirectory
*	Save Backup Data to File (Toggle on/off)
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
char GetBCommand(int backup, char **prompt, int *resource)
{
	char command;
	int x, y;

	*prompt = "resource";
	*resource = TRUE;

	clrscr();
	printf(	"T:  %save backup data\n"
			"BR: Backup Resource\n"
			"BS: Backup Subdirectories\n"
			"SR: Scan Resource\n"
			"SS: Scan Subdirectories\n\n"
			"Q:  Return to previous menu\n"
			"\nSelect option: ", backup ? "S" : "Do not s");
	y = wherey(), x = wherex();


	loop
	{
		gotoxy(x, y);
		printf("  ");
		gotoxy(x, y);

		switch (command = toupper(getch()))
		{
		case 'Q':
		case 'T':
			goto Return;

		case 'S':
		case 'B':
			putch(command);
			switch (toupper(getch()))
			{
			case 'S':
				*prompt = "subdirectory";
				*resource = FALSE;

			case 'R':
				goto Return;
			}

		default:
			putch('\a');
			break;
		}
	}

Return:

	while (kbhit()) getch();
	clrscr();
	return (command);
}
/****************************************************************************/

/*********************************************************************
* NAME: GetRCommand
*
* WHAT: Prompt the user to select the following "Restore" options:
*
*	Overwrite/Don't Overwrite Parent
*	Overwrite/Don't Overwrite Child
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
char GetRCommand(UINT8 overwriteParent, UINT8 overwriteChild)
{
	char command;
	int x, y;

	clrscr();
	printf( "R: Restore Session\n"
			"P: %sverwrite Parent\n"
			"C: %sverwrite Child\n\n"
			"Q: Return to previous menu\n"
			"\nSelect option: ", overwriteParent ? "O" : "Do not o", overwriteChild ? "O" : "Do not o");
	y = wherey(), x = wherex();


	loop
	{
		gotoxy(x, y);
		printf("  ");
		gotoxy(x, y);

		command = toupper(getch());
		if (strchr("RPCQ", command))
			goto Return;

		putch('\a');
	}

Return:

	while (kbhit()) getch();
	clrscr();
	return (command);
}
/****************************************************************************/

STRING colon = ":", colonColon = "::", slash = "/";

/*********************************************************************
* NAME: GetNameSpaceType
*
* WHAT: Prompt the user to select the Data Set Name Space Type. Verify
*	that the selected type is valid. The name space type is used
*	in scanning data sets on a Target Service.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
UINT32 GetNameSpaceType(void)
{
	char option;
	int nameSpaceCount;
	int realIndex, subIndex;
	int x, y;

	UINT32 nameSpaceType = NOT_ASSIGNED;
	UINT32 sequence = 0;
	NWSM_NAME_LIST *supportedNameSpaces = NULL;
	NWSM_NAME_LIST *saveFirstNameSpace = NULL;
	CCODE ccode = 0;
	BUFFER resourceName[256];

	

	ccode=NWSMTSScanTargetServiceResource(connection, &sequence, resourceName);

	/************************************************************************************/
	/* we will get all resources here (sequence = 0) so that supported name spaces      */
	/*	will list all supported name spaces available on the system.                     */
	/************************************************************************************/

	if (ccode != 0)
		{
		printf("\t\aSomethings wrong... I quit!");
		nameSpaceType=100;
		goto NameSpaceEnd;
		}

	ccode=NWSMTSListSupportedNameSpaces(connection, resourceName, &supportedNameSpaces);

	if (ccode == NWSMTS_OUT_OF_MEMORY)
		{
		printf("\t\aOut of Memory\n");
		pause();
		nameSpaceType=100;
		goto NameSpaceEnd;
		}
	saveFirstNameSpace = supportedNameSpaces;

	/**************************************************************************/
	/* This routine is based on the idea that at least one name space will be */
	/* returned from the TSA.                                                 */
	/**************************************************************************/

	nameSpaceCount = 1;
	printf("\n\nName Space Type:\n");
	while (supportedNameSpaces != NULL)
		{
		printf("\t%d: %s\n",nameSpaceCount, supportedNameSpaces->name+4);
		supportedNameSpaces = supportedNameSpaces->next;
		nameSpaceCount++;
	
		}
	printf("\t%d: Quit\n", nameSpaceCount);

	printf("\n\nEnter option: ");
	x=wherex();
	y=wherey();
	option = getch();
	realIndex = option - '0';
	if (realIndex == nameSpaceCount)
		{
		nameSpaceType=100;
		goto NameSpaceEnd;
		}
	while((realIndex < 1) || (realIndex > nameSpaceCount))
		{
		printf("\a\a");
		gotoxy(x,y);
		option = getch();
		realIndex = option - '0';
		if (realIndex == nameSpaceCount)
			{
			nameSpaceType=100;
			goto NameSpaceEnd;
			}

		}

	supportedNameSpaces = saveFirstNameSpace;
	for (subIndex=1; subIndex < realIndex; subIndex++)
		{
		supportedNameSpaces = supportedNameSpaces->next;
		}
	nameSpaceType=*(UINT32 *)supportedNameSpaces->name;

/*** nameSpaceType is what we're passing back  ***/

	
NameSpaceEnd:
	NWSMFreeNameList(&saveFirstNameSpace);
	saveFirstNameSpace = NULL;
	return (nameSpaceType);

}
/****************************************************************************/

/*********************************************************************
* NAME: GetSelectionType
*
* WHAT: Prompt the user to enter the type of selection desired for the
*	data set scan. 
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
UINT32 GetSelectionType(void)
{
	char *strings[62]; /*This builds an array of pointers to string1 and string2 */
	char string1[128];
	char string2[128];
	char option;
	int x, y;
	int index=1, endindex=1;
	UINT8 typenum=1;
	UINT32 selectionType = NOT_ASSIGNED;
	UINT32 value=2;
	UINT32 retValue[62];
	CCODE ccode=0;

	while (ccode == 0)
		{
		ccode=NWSMTSGetTargetSelectionTypeStr(connection, typenum, string1, string2);
		if (ccode == NWSMTS_INVALID_CONNECTION_HANDL)
			{
			printf("\nINVALID CONNECTION HANDLE");
			break;
			}

		if (ccode == NWSMTS_INVALID_SELECTION_TYPE)
			{
			printf("\nINVALID SELECTION TYPE");
			break;
			}
		if (*string1 != NULL)
			{
			strings[index]=malloc(strlen(string1)+1);
			strcpy(strings[index], string1);
			retValue[index]=value;
			index++;
			}

		if (*string2 != NULL)
			{
			strings[index]=malloc(strlen(string2)+1);
			strcpy(strings[index], string2);
			retValue[index]=(value|1);
			index++;
			}
		typenum++;		
		value <<= 1;
		}


	endindex=index;
	printf("\n\nSelection Type:\n");
	for (index = 1; index < endindex; index++)
		{
		if (strings[index] != NULL)
			{
			printf("\t%d: %s\n",index, strings[index]);
			}else printf("BIG GOOF HERE\a\a\a");
		}
	printf("\t%d: Return to previous menu\n", endindex);
	printf("\n\nEnter Selection Type: ");


	y = wherey(), x = wherex();


	while (selectionType is NOT_ASSIGNED)
	{
		gotoxy(x, y);
		printf("  ");
		gotoxy(x, y);

		option = getch();
		index = option - '0';
		
		if ((index < 1) || (index > endindex))
			{
			printf("\a\a");
			selectionType = NOT_ASSIGNED;
			continue;
			}

		if (index == endindex)
			{
			putch(option);
			selectionType=100;
			break;
			}
		putch(option);
		selectionType=retValue[index];



	}


	for (index=1; index < endindex; index++)
		{
		if (strings[index] != NULL)
			free(strings[index]);
		}
 	return (selectionType);
}
/****************************************************************************/

/*********************************************************************
* NAME: SetNameSpaceType
*
* WHAT: Convert a name space type code (UINT32) into a string describing
*	the name space type passed as a parameter to this function.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
void SetNameSpaceType(UINT32 selectedNameSpaceType, char *string)
{
	UINT32 sequence = 0;
	NWSM_NAME_LIST *supportedNameSpaces = NULL; /** I just added this for destruction of hard-coded junk 12/9/91 gn **/
	NWSM_NAME_LIST *saveFirstNameSpace = NULL;
	CCODE ccode = 0;
	BUFFER resourceName[256];


	ccode=NWSMTSScanTargetServiceResource(connection, &sequence, resourceName);

		if (ccode != 0)
		{
		printf("\t\aSomethings wrong... I quit!");
		strcpy(string, "Bad nameSpaceType");
		goto GetStringEnd;
		}

	ccode=NWSMTSListSupportedNameSpaces(connection, resourceName, &supportedNameSpaces);

	if (ccode == NWSMTS_OUT_OF_MEMORY)
		{
		printf("\t\aOut of Memory\n");
		pause();
		strcpy(string, "Bad nameSpaceType");
		goto GetStringEnd;
		}
	saveFirstNameSpace = supportedNameSpaces;

	/**************************************************************************/
	/* This routine is based on the idea that at least one name space will be */
	/* returned from the TSA.                                                 */
	/**************************************************************************/

	while (*supportedNameSpaces->name != selectedNameSpaceType)
		{
		supportedNameSpaces = supportedNameSpaces->next;
		if (supportedNameSpaces == NULL)
			break;
	
		}

	if (*supportedNameSpaces->name != selectedNameSpaceType)
		{
		strcpy(string, "Bad nameSpaceType");
		goto GetStringEnd;
		}
	strcpy(string, (char *)supportedNameSpaces->name+4);

/*** string is what we're passing back  ***/

	
GetStringEnd:

	NWSMFreeNameList(&saveFirstNameSpace);
	saveFirstNameSpace = NULL;
	return;
}
/****************************************************************************/

/*********************************************************************
* NAME: SetSelectionType
*
* WHAT: Convert a selection type code (UINT32) into a string describing
*	the selection type passed as a parameter to this function.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
void SetSelectionType(UINT32 selectionType, char *string)
{
	int index=1;
	UINT8 typenum=1;
	char string1[128], string2[128];
	UINT32 value=2;
	CCODE ccode=0;

	for (index = 1, value=2; index <= 32; index++, value <<= 1)
		{
		ccode=NWSMTSGetTargetSelectionTypeStr(connection, typenum, string1, string2);
		if (ccode == NWSMTS_INVALID_CONNECTION_HANDL)
			{
			printf("\nINVALID CONNECTION HANDLE");
			break;
			}

		if (ccode == NWSMTS_INVALID_SELECTION_TYPE)
			{
			printf("\nINVALID SELECTION TYPE");
			continue;
			}
		if (value == selectionType)
			{
			strcpy(string, (char *)string1);
			break;
			}

		if ((value|1) == selectionType)
			{
			strcpy(string, (char *)string2);
			break;
			}

		typenum++;
		}

	return;
/*	STRING string = "Bad selectionType";*/
/**/
/*	switch ((UINT16)selectionType)*/
/*	{*/
/*	case 2:*/
/*	 	string = "Exclude File Server Level";*/
/*		break;*/
/**/
/*	case 4:*/
/*	 	string = "Exclude Directories";*/
/*		break;*/
/**/
/*	case 5:*/
/*	 	string = "Include Directories";*/
/*		break;*/
/**/
/*	case 8:*/
/*	 	string = "Exclude Files";*/
/*		break;*/
/**/
/*	case 9:*/
/*	 	string = "Include Files";*/
/*		break;*/
/**/
/*	case 16:*/
/*	 	string = "Exclude Path/Files";*/
/*		break;*/
/**/
/*	case 17:*/
/*	 	string = "Include Path/Files";*/
/*		break;*/
/*	}*/

}
/****************************************************************************/

static FILE *dataFile=NULL, *indexFile=NULL;
extern BUFFER dataBuffer[];
struct
{
	UINT8  parent;
	UINT8  terminalNodeNameOnly;
	UINT32 recordSize;
} DEdata;
/****************************************************************************/

/*********************************************************************
* NAME: WriteDEFiles
*
* WHAT: Write backup data to disk. For the demonstration engine, backup
*	data is placed in DOS files instead of written to tape (or other
*	backup media). Two files are created and updated - an index file
*	and a file that actually contains the backup data. Each data set
*	is separated by an END_OF_RECORD_MARKER that has a value of
*	0xFFFFFFFF in the size field of the block (the size is kept in
*	the index file). Data sets larger than 50,000 bytes are split into
*	blocks less than or equal to 50,000 bytes.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
CCODE WriteDEFiles(BUFFER *buffer, UINT32 size, UINT8 parent, UINT8 terminal)
{
	CCODE ccode = 0;

	if (!indexFile or !dataFile)
		ccode = 1;

	else if (size is END_OF_RECORD_MARKER)		/* write EOF record */
	{
		DEdata.parent = 0;
		DEdata.terminalNodeNameOnly = 0;
		DEdata.recordSize = size;
		if (fwrite(&DEdata, sizeof(DEdata), 1, indexFile) isnt 1)
		{
			perror("\n\nWriteDEFiles error writing index file (END_OF_RECORD_MARKER)");
			ccode = 1;
		}
	}

	else
	{
		DEdata.parent = parent;
		DEdata.terminalNodeNameOnly = terminal;
		DEdata.recordSize = size;
		if (fwrite(&DEdata, sizeof(DEdata), 1, indexFile) isnt 1)
		{
			perror("\n\nWriteDEFiles error writing index file");
			ccode = 1;
		}

		if (fwrite(buffer, sizeof(char), (size_t)size, dataFile) isnt size)
		{
			perror("\n\nWriteDEFiles error writing data file");
			ccode = 1;
		}
	}

	return (ccode);
}
/****************************************************************************/

static int ungetEOFFlag = FALSE;
void UngetEOF(void)
{
	ungetEOFFlag = TRUE;
}

/*********************************************************************
* NAME: ReadDEFiles
*
* WHAT: Read the data that was stored on disk during a backup session.
*	See the comments in the header for the WriteDEFiles function for
*	more details.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
BUFFER *ReadDEFiles(UINT32 *size, UINT8 *parentFlag, UINT8 *tnoFlag)
{
	BUFFER *buffer = NULL;
	*size = 0;

	if (ungetEOFFlag)
	{
		ungetEOFFlag = FALSE;
		DEdata.recordSize = END_OF_RECORD_MARKER;
		*size = DEdata.recordSize;
		buffer = dataBuffer;
		goto Return;
	}

/*	read index file */
	if (fread((char *)&DEdata, sizeof(DEdata), 1, indexFile) isnt 1)
	{
		if (!feof(indexFile))
			perror("\n\nReadDEFiles error reading index file");
		goto Return;
	}

/*	read data file */
	if (DEdata.recordSize isnt END_OF_RECORD_MARKER)	/* normal data record */
	{
		if (fread((void *)dataBuffer, sizeof(char), (size_t )DEdata.recordSize, dataFile) isnt DEdata.recordSize)
		{
			if (!feof(dataFile))
				perror("\n\nReadDEFiles error reading data file");
			goto Return;
		}
	}

	*size = DEdata.recordSize;
	*parentFlag = DEdata.parent;
	*tnoFlag = DEdata.terminalNodeNameOnly;
	buffer = dataBuffer;

Return:
	return (buffer);
}
/****************************************************************************/

/*********************************************************************
* NAME: OpenDEFiles
*
* WHAT: Open an index file and a data file for storing backup data from
*	a Target Service.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
CCODE OpenDEFiles(char mode)
{
	BUFFER flags[3];
	CCODE  ccode = 0;
	FILE  *fopen();

	sprintf(flags, "%cb", tolower(mode));
	if ((dataFile = fopen("de-exmpl.dat", flags)) is NULL)
	{
		perror("\n\nData file open error");
		ccode = 1;
	}

	if ((indexFile = fopen("de-exmpl.idx", flags)) is NULL)
	{
		perror("\n\nIndex file open error");
		ccode = 1;
	}

	if (ccode)
		CloseDEFiles();

	else if (mode is 'w')
		WriteDEFiles(NULL, END_OF_RECORD_MARKER, 0, 0);

	return (ccode);
}
/****************************************************************************/

/*********************************************************************
* NAME: CloseDEFiles
*
* WHAT: Close the index and data files opened with OpenDEFiles.
*
* SMS-SPECIFIC ROUTINES CALLED:
*
**********************************************************************/
void CloseDEFiles(void)
{
	if (dataFile and fclose(dataFile))
		perror("\n\nData file close error");

	if (indexFile and fclose(indexFile))
		perror("\n\nIndex file close error");

	dataFile = indexFile = NULL;
}

/****************************************************************************/
/****************************************************************************/

