/*********************************************************************
* NAME: ProcessErrorLog
*
* WHAT: Fetchs and displays "ERROR LOG" data set from the Target Service.
*
* SMS-SPECIFIC ROUTINES CALLED: 
*				NWSMPutOneName (SMS UTILITY API)
*				NWSMTSScanDataSetBegin (TSAPI)
*				NWSMTSOpenDataSetForBackup (TSAPI)
*				NWSMTSReadDataSet (TSAPI)
*				NWSMTSCloseDataSet (TSAPI)
*				NWSMTSScanDataSetEnd (TSAPI)
*
**********************************************************************/
void ProcessErrorLog(void)
{
  /* connection is global value in this example */
  /* dataBuffer is global value in this example */
  CCODE                      ccode;
  NWSM_DATA_SET_NAME_LIST   *resourceName = NULL;
  NWSM_DATA_SET_NAME_LIST   *dataSetNameList = NULL;
  UINT32                     sequence;
  UINT32                     mode = NWSM_USE_LOCK_MODE_IF_DW_FAILS;
  UINT32                     bytesRead;
  UINT32                     dataHandle = NULL;
  UINT16                     readSize = BUFFER_SIZE;
  int                        empty = TRUE;

  /* display error log name */
  printf ("\n\n\"ERROR LOG\":\n");

  /* setup resource name list */
  if ((ccode = NWSMPutOneName( 
                          &resourceName, 
                          NWSM_TSA_DEFINED_RESOURCE_TYPE,
                          0L,                              /* selectionType */
                          FALSE,                           /* reverse order */
                          NULL,                          /* first separator */
                          NULL,                         /* second separator */
                          "ERROR LOG" ))          /* Error log special name */
       != 0 )
  {
    PrintError( 
      ccode, 
      "DE: * * *   ERROR  * * * NWSMPutOneName(ERROR LOG)", 
      0);
    goto Terminate;
  }

  /* read in error log data set */
  if ((ccode = NWSMTSScanDataSetBegin( 
                                 connection,
                                 resourceName,
                                 NULL,                      /* scan control */
                                 NULL,                    /* selection list */
                                 &sequence,
                                 NULL,                  /* scan information */
                                 &dataSetNameList ))
       != 0 )
  {
    PrintError( 
      ccode, 
      "DE: * * *   ERROR   * * * NWSMTSScanDataSetBegin(ERROR LOG)", 
      0);
    goto Terminate;
  }
  if (((ccode = NWSMTSOpenDataSetForBackup( 
                                          connection,
                                          sequence,
                                          0L,                       /* mode */
                                          &dataHandle ))
       !=0 ) || ( dataHandle == 0L ) )
  {
    PrintError( 
      ccode, 
      "DE: * * *   ERROR   * * * NWSMTSOpenDataSetForBackup(ERROR LOG)", 
      0);
    goto Terminate;
  }
  do
  {
    if (( ccode = NWSMTSReadDataSet( 
                               connection,
                               dataHandle,
                               readSize,
                               &bytesRead,
                               dataBuffer ))
          != 0 )
    {
      PrintError( 
        ccode, 
        "DE: * * *   ERROR   * * * NWSMTSReadDataSet(ERROR LOG)", 
        0);
      break;
    }
    if (bytesRead)
    {
      dataBuffer[ bytesRead ] = 0;
      printf( dataBuffer );
      empty = FALSE;
    }
    if ( bytesRead < readSize )
    {
      break;
    }
  } while ( ccode == 0 );
Terminate:

  if (empty)
  {
    printf( "\t<empty>\n");
  }

  if ( dataHandle )
  {
    if (( ccode = NWSMTSCloseDataSet(
                                connection,
                                &dataHandle ))
          != 0 && ccode != NWSMTS_INVALID_DATA_SET_HANDLE )
    {
      PrintError( 
        ccode, 
        "DE: * * *   ERROR   * * * NWSMTSCloseDataSet(ERROR LOG)", 
        0);
    }
  }
  if (( ccode = NWSMTSScanDataSetEnd( 
                                connection,
                                &sequence,
                                NULL, /* scan information */
                                &dataSetNameList ))
        != 0 )
  {
    PrintError( 
      ccode, 
      "DE: * * *   ERROR   * * * NWSMTSScanDataSetEnd(ERROR LOG)", 
      0);
  }
  if ( resourceName )
    free( resourceName );
  if ( dataSetNameList )
    free( dataSetNameList );

} /* ProcessErrorLog() */

/****************************************************************************/